"""Dashboard objects."""
from typing import Any
from datetime import datetime

import polars as pl
import panel as pn
import pandas as pd
from panel.template import MaterialTemplate
from panel.viewable import Viewer

from nwm_explorer.nwm import nwm_site_generator
from nwm_explorer.evaluate import load_metrics, scan_evaluations, load_site_metrics
from nwm_explorer.routelink import download_routelink
from nwm_explorer.usgs import usgs_site_generator, load_site_information
from nwm_explorer.views import (FilterWidgets, MapView, TimeSeriesView, BarPlot, ECDFMatrix,
    MarkdownView, ECDFSelector)
from nwm_explorer.constants import (METRIC_PLOTTING_LIMITS, CONFIGURATION_LINE_TYPE, SITE_COLUMN_MAPPING,
    MeasurementUnits, ModelConfiguration)
from nwm_explorer.options import StreamflowOptions, compute_conversion_factor
from nwm_explorer.configuration import Configuration

def generate_usgs_url(
        usgs_site_code: str,
        start_datetime: datetime,
        end_datetime: datetime
    ) -> str:
    """Returns URL to USGS monitoring location page."""
    # Base monitoring location URL
    url = f"https://waterdata.usgs.gov/monitoring-location/USGS-{usgs_site_code}/#"

    # parameters
    params = {
        "dataTypeId": "continuous-00060-0",
        "showFieldMeasurements": "true",
        "startDT": start_datetime.strftime("%Y-%m-%dT%H:%M"),
        "endDT": end_datetime.strftime("%Y-%m-%dT%H:%M")
    }

    # Finish building URL
    for k, v in params.items():
        url += f"&{k}={v}"
    return url

class Dashboard(Viewer):
    """Build a dashboard for exploring National Water Model output."""
    def __init__(self, configuration: Configuration, **params):
        super().__init__(**params)
        # Handle configuration
        root = configuration.root
        title = configuration.title

        # Open routelink
        routelink = download_routelink(root).select(
            ["nwm_feature_id", "latitude", "longitude"]
        ).collect()

        filter_widgets = FilterWidgets(
            evaluation_options=scan_evaluations(root, cache=True).select(
                "label").collect().unique()["label"].to_list()
        )
        data_ranges: dict[str, pd.Timestamp] = {
            "observed_value_time_min": None,
            "observed_value_time_max": None,
            "reference_time_min": None,
            "reference_time_max": None,
        }
        self.usgs_site_code: str | None = None
        self.nwm_feature_id: int | None = None

        site_map = MapView(map_layers=configuration.map_layers)
        hydrograph = TimeSeriesView()
        barplot = BarPlot()
        ecdf = ECDFMatrix(nplots=4, ncols=2)
        ecdf_filters = ECDFSelector(nplots=4, filter_widgets=filter_widgets)
        site_information = MarkdownView()
        streamflow_options = StreamflowOptions()

        def draw_hydrograph(event, callback_type: str) -> None:
            # Ignore non-event
            if event is None:
                return

            # Ignore axis scale change
            if callback_type in ["axis_scale"]:
                return

            # Check for site selection
            if self.usgs_site_code is None:
                return

            # Clear hydrograph
            hydrograph.erase_data(
                xrange=(
                    data_ranges["observed_value_time_min"],
                    data_ranges["observed_value_time_max"]
                    )
            )

            # Unit conversion
            conversion_factor = compute_conversion_factor(
                root, self.usgs_site_code, streamflow_options.measurement_units
            )

            # Update y-axis title
            hydrograph.set_yaxis_title(f"Streamflow ({streamflow_options.measurement_units})")

            # Stream observations
            dataframes = []
            for df in usgs_site_generator(
                root=root,
                usgs_site_code=self.usgs_site_code,
                start_time=data_ranges["observed_value_time_min"],
                end_time=data_ranges["observed_value_time_max"],
                cache=True
            ):
                # Check for data
                if df.is_empty():
                    continue

                # Append data
                dataframes.append(df)
                observations = pl.concat(dataframes)

                # Resample
                if streamflow_options.measurement_units in [
                    MeasurementUnits.INCHES_PER_HOUR,
                    MeasurementUnits.CUMULATIVE_INCHES_PER_HOUR
                ]:
                    observations = observations.group_by_dynamic(
                        "value_time", every="1h"
                    ).agg(
                        pl.col("observed_cfs").mean()
                    ).upsample(
                        time_column="value_time", every="1h"
                    ).fill_null(
                        strategy="forward"
                    )

                # Apply conversion
                if conversion_factor != 1.0:
                    observations = observations.with_columns(
                        pl.col("observed_cfs").mul(conversion_factor)
                    )

                # Accumulate
                if streamflow_options.measurement_units in [
                    MeasurementUnits.CUMULATIVE_INCHES_PER_HOUR
                ]:
                    observations = observations.with_columns(
                        pl.col("observed_cfs").cum_sum()
                    )

                # Replace data
                hydrograph.update_trace(
                    xdata=observations["value_time"].to_numpy(),
                    ydata=observations["observed_cfs"].to_numpy(),
                    name=f"USGS-{self.usgs_site_code}"
                )

            # Stream predictions
            accumulated_value = 0.0
            for df in nwm_site_generator(
                root=root,
                configuration=filter_widgets.configuration,
                nwm_feature_id=self.nwm_feature_id,
                start_time=data_ranges["reference_time_min"],
                end_time=data_ranges["reference_time_max"],
                cache=True
            ):
                # Add date string column
                df = df.with_columns(
                    pl.col(
                        "reference_time"
                    ).dt.strftime(
                        "Issued: %Y-%m-%d %HZ"
                    ).alias("datetime_string")
                ).sort(["reference_time", "value_time"])

                # Add each reference time
                trace_data = []
                for rt in df["datetime_string"].unique(maintain_order=True):
                    # Extract predictions
                    predictions = df.filter(pl.col("datetime_string") == rt)

                    # Apply conversion
                    if conversion_factor != 1.0:
                        predictions = predictions.with_columns(
                            pl.col("predicted_cfs").mul(conversion_factor)
                        )

                    # Accumulate
                    if streamflow_options.measurement_units in [
                        MeasurementUnits.CUMULATIVE_INCHES_PER_HOUR
                    ]:
                        # Distinguish between analysis and forecast
                        if filter_widgets.configuration in [
                            ModelConfiguration.ANALYSIS_ASSIM_EXTEND_ALASKA_NO_DA,
                            ModelConfiguration.ANALYSIS_ASSIM_EXTEND_NO_DA,
                            ModelConfiguration.ANALYSIS_ASSIM_HAWAII_NO_DA,
                            ModelConfiguration.ANALYSIS_ASSIM_PUERTO_RICO_NO_DA
                        ]:
                            predictions = predictions.with_columns(
                                pl.col("predicted_cfs").cum_sum().add(accumulated_value)
                            )
                            accumulated_value = predictions["predicted_cfs"].max()
                        else:
                            intial_time = predictions["value_time"].min() - pl.duration(hours=1)
                            vals = observations.filter(
                                pl.col("value_time") == intial_time
                            )["observed_cfs"]
                            if vals.is_empty():
                                accumulated_value = 0.0
                            else:
                                accumulated_value = vals.item(0)
                            predictions = predictions.with_columns(
                                pl.col("predicted_cfs").cum_sum().add(accumulated_value)
                            )

                    # Add trace data
                    trace_data.append((
                        predictions["value_time"].to_numpy(),
                        predictions["predicted_cfs"].to_numpy(),
                        rt
                    ))

                # Add to plot
                mode = CONFIGURATION_LINE_TYPE.get(
                    filter_widgets.configuration, "lines"
                )
                hydrograph.append_traces(trace_data, mode=mode)

        def handle_map_click(event, callback_type: str) -> None:
            if event is None:
                return

            # Check for state
            if not site_map.click_data:
                # Clear barplot
                barplot.erase()

                # Clear hydrograph
                hydrograph.erase_data(
                    xrange=(
                        data_ranges["observed_value_time_min"],
                        data_ranges["observed_value_time_max"]
                        )
                )

                # Clear site info
                site_information.erase()
                self.usgs_site_code = None
                self.nwm_feature_id = None
                return

            # Parse custom data
            metadata = site_map.click_data["customdata"]
            self.nwm_feature_id = metadata[0]
            self.usgs_site_code = metadata[1]

            # Update site info
            if callback_type == "click":
                # Build site table
                info = "| Site Information |  |  \n| :-- | :-- |  \n"
                for series in load_site_information(root, self.usgs_site_code,
                    rename=SITE_COLUMN_MAPPING).iter_columns():
                    info += f"| **{series.name}** | {series.item(0)} |  \n"
                url = generate_usgs_url(
                    self.usgs_site_code,
                    data_ranges["observed_value_time_min"],
                    data_ranges["observed_value_time_max"]
                    )
                info += "| **Monitoring page** | "
                info += f'<a href="{url}" target="_blank">Open new tab</a> |'
                site_information.update(info)

            # These filters do not update hydrograph or barplot
            if callback_type in ["lead_time", "significant"]:
                return

            # Retrieve metrics
            metric_data = load_site_metrics(
                root=root,
                label=filter_widgets.label,
                configuration=filter_widgets.configuration,
                metric=filter_widgets.metric,
                nwm_feature_id=self.nwm_feature_id,
                rank=filter_widgets.rank,
                cache=True
            )

            # Clear barplot
            barplot.erase()

            # Update barplot
            barplot.update(
                xdata=metric_data["lead_time_hours_min"],
                ydata=metric_data[filter_widgets.point_column].to_numpy(),
                ydata_lower=metric_data[filter_widgets.lower_column].to_numpy(),
                ydata_upper=metric_data[filter_widgets.upper_column].to_numpy(),
                xlabel="Minimum Lead Time (h)",
                ylabel=filter_widgets.metric_label
            )

            # These filters do not update hydrograph
            if callback_type in ["rank", "metric"]:
                return

            # Update hydrograph
            draw_hydrograph(event, callback_type)
        site_map.bind_click(handle_map_click)

        def handle_relayout(event: dict[str, Any], callback_type: str) -> None:
            # Ignore non-calls
            if event is None or callback_type is None:
                return

            # Update each plot
            for p in ecdf_filters:
                # Retrieve data
                data = load_metrics(
                    root=root,
                    label=filter_widgets.label,
                    configuration=filter_widgets.configuration,
                    metric=p.metric,
                    lead_time_hours_min=filter_widgets.lead_time,
                    rank=filter_widgets.rank,
                    additional_columns=(
                        "nwm_feature_id",
                        "usgs_site_code",
                        "sample_size"
                        ),
                    significant=filter_widgets.significant,
                    cache=True
                ).with_columns(
                    latitude=pl.col("nwm_feature_id").replace_strict(
                        old=routelink["nwm_feature_id"].implode(),
                        new=routelink["latitude"].implode()
                    ),
                    longitude=pl.col("nwm_feature_id").replace_strict(
                        old=routelink["nwm_feature_id"].implode(),
                        new=routelink["longitude"].implode()
                    )
                )

                # Sort data
                if site_map.viewport:
                    sorted_data = data.filter(
                        pl.col("latitude") <= site_map.viewport["lat_max"],
                        pl.col("latitude") >= site_map.viewport["lat_min"],
                        pl.col("longitude") <= site_map.viewport["lon_max"],
                        pl.col("longitude") >= site_map.viewport["lon_min"]
                    ).sort(p.point_column, descending=False)
                else:
                    sorted_data = data.sort(p.point_column, descending=False)

                # Update ECDF
                ecdf.update(
                    index=p.index,
                    xdata=sorted_data[p.point_column].to_numpy(),
                    xdata_lower=sorted_data[p.lower_column].to_numpy(),
                    xdata_upper=sorted_data[p.upper_column].to_numpy(),
                    xlabel=p.metric_label,
                    xrange=METRIC_PLOTTING_LIMITS[p.metric]
                )
        site_map.bind_relayout(handle_relayout)
        ecdf_filters.bind(handle_relayout)

        def handle_filter_updates(event: str, callback_type: str) -> None:
            # Ignore non-calls
            if event is None:
                return

            # Retrieve data
            data = load_metrics(
                root=root,
                label=filter_widgets.label,
                configuration=filter_widgets.configuration,
                metric=filter_widgets.metric,
                lead_time_hours_min=filter_widgets.lead_time,
                rank=filter_widgets.rank,
                additional_columns=(
                    "nwm_feature_id",
                    "usgs_site_code",
                    "sample_size",
                    "observed_value_time_min",
                    "observed_value_time_max",
                    "reference_time_min",
                    "reference_time_max"
                    ),
                significant=filter_widgets.significant,
                cache=True
            ).with_columns(
                latitude=pl.col("nwm_feature_id").replace_strict(
                    old=routelink["nwm_feature_id"].implode(),
                    new=routelink["latitude"].implode()
                ),
                longitude=pl.col("nwm_feature_id").replace_strict(
                    old=routelink["nwm_feature_id"].implode(),
                    new=routelink["longitude"].implode()
                )
            )

            # Update date range
            data_ranges["observed_value_time_min"] = data["observed_value_time_min"].min()
            data_ranges["observed_value_time_max"] = data["observed_value_time_max"].max()
            data_ranges["reference_time_min"] = data["reference_time_min"].min()
            data_ranges["reference_time_max"] = data["reference_time_max"].max()

            # Update map
            site_map.update(
                dataframe=data,
                column=filter_widgets.point_column,
                domain=filter_widgets.domain,
                cmin=METRIC_PLOTTING_LIMITS[filter_widgets.metric][0],
                cmax=METRIC_PLOTTING_LIMITS[filter_widgets.metric][1],
                metric_label=filter_widgets.metric_label,
                custom_data=data.to_pandas()[[
                        "nwm_feature_id",
                        "usgs_site_code",
                        filter_widgets.lower_column,
                        filter_widgets.upper_column,
                        "sample_size"
                    ]],
                hover_template=(
                        f"{filter_widgets.metric_label}: "
                        "%{marker.color:.2f}<br>"
                        "95% CI: %{customdata[2]:.2f} -- %{customdata[3]:.2f}<br>"
                        "Samples: %{customdata[4]:.0f}<br>"
                        "NWM Feature ID: %{customdata[0]}<br>"
                        "USGS Site Code: %{customdata[1]}<br>"
                        "Longitude: %{lon}<br>"
                        "Latitude: %{lat}"
                    )
                )

            # Update ECDF
            handle_relayout(0, callback_type)

            # Update barplot and hydrograph
            handle_map_click(0, callback_type)
        filter_widgets.bind(handle_filter_updates)

        # Handle streamflow options
        def handle_streamflow_options(event: str, callback_type: str) -> None:
            if event is None:
                return

            # Change scale
            if callback_type == "axis_scale":
                hydrograph.set_axis_type(streamflow_options.axis_scale)
                return

            # Apply options (i.e. change measurement units)
            draw_hydrograph(event, callback_type)
        streamflow_options.bind(handle_streamflow_options)

        # Initialize states
        handle_filter_updates(filter_widgets.label, "intialize")
        handle_relayout(0, "intialize")

        # Setup template
        self.template = MaterialTemplate(
            title=title,
            collapsed_sidebar=True
        )

        # Main area
        controls = pn.Column(filter_widgets, site_information)
        content = pn.Column(
            pn.Row(site_map, ecdf),
            pn.Row(hydrograph, barplot)
        )
        self.template.main.append(pn.Row(
            controls,
            content
        ))

        # Sidebar
        self.template.sidebar.append(ecdf_filters)
        self.template.sidebar.append(streamflow_options)
        self.template.sidebar.append(site_map.map_layer_selector)

    def __panel__(self) -> MaterialTemplate:
        return self.template



"""Utilities to support downloading files."""
from typing import Callable
from time import sleep
from pathlib import Path
import asyncio
import ssl
from http import HTTPStatus
import warnings
import inspect
from yarl import URL
import aiohttp
from aiohttp.typedefs import LooseHeaders
import aiofiles

from nwm_explorer.logger import get_logger

class FileValidationError(Exception):
    """Exception raised when file fails to validate."""

def default_file_validator(filepath: Path) -> None:
    """
    Validate that given filepath opens and closes without raising.

    Parameters
    ----------
    filepath: Path
        Path to file.
    
    Returns
    -------
    None

    Raises
    ------
    FileValidationError
    """
    try:
        with filepath.open("rb"):
            return
    except Exception as e:
        raise FileValidationError(e) from e

async def download_file_awaitable(
        url: str | URL,
        filepath: str | Path,
        session: aiohttp.ClientSession,
        chunk_size: int = 1024,
        overwrite: bool = False,
        ssl_context: ssl.SSLContext | None = None,
        timeout: int = 300,
        retries: int = 3
        ) -> None:
    """
    Retrieve a single file from url and save to filepath.
    
    Parameters
    ----------
    url: str | URL, required
        Source URL to download.
    filepath: str | Path, required
        Destination to save file.
    session: ClientSession, required
        Session object used for retrieval.
    chunk_size: int, optional, default 1024
        Amount of data to write at a time (KB).
    overwrite: bool, optional, default False
        If filepath exists, overwrite if True, else skip download.
    ssl_context: SSLContext, optional
        SSL configuration object. Uses system defaults unless otherwise
        specified.
    timeout: int, optional, default 300
        Maximum number of seconds to wait for a response to return.
    retries: int, optional, default 3
        Number of download attempts in the event of failure.
        
    Returns
    -------
    None
    """
    # Check for file existence
    if not overwrite and Path(filepath).exists():
        message = f"File exists, skipping download of {filepath}"
        warnings.warn(message, UserWarning)
        return

    # SSL
    if ssl_context is None:
        ssl_context = ssl.create_default_context()

    # Fetch
    for attempt in range(retries):
        try:
            async with session.get(url, ssl=ssl_context, timeout=timeout,
                raise_for_status=True) as response:
                # Stream download
                async with aiofiles.open(filepath, 'wb') as fo:
                    while True:
                        chunk = await response.content.read(chunk_size)
                        if not chunk:
                            break
                        await fo.write(chunk)
                return
        except aiohttp.ClientResponseError as e:
            status = HTTPStatus(e.status)
            message = (
                f"HTTP Status: {status.value}" +
                f" - {status.phrase}" +
                f" - {status.description}\n" +
                f"{url}"
                )
            warnings.warn(message, RuntimeWarning)

        # Sleep
        await asyncio.sleep(5 * 2 ** attempt)

    # Unable to retrieve file
    message = f"Unable to retrieve {url}"
    warnings.warn(message, RuntimeWarning)

async def download_files_awaitable(
        *src_dst: tuple[str | URL, str | Path],
        auto_decompress: bool = True,
        headers: LooseHeaders | None = None,
        limit: int = 100,
        chunk_size: int = 1024,
        overwrite: bool = False,
        ssl_context: ssl.SSLContext | None = None,
        timeout: int = 300,
        retries: int = 3
        ) -> None:
    """
    Asynchronously retrieve multiple files from urls and save to filepaths.
    
    Parameters
    ----------
    *src_dst: tuple[str | Path, str | URL], required
        One or more tuples containing two values. The first value is the 
        source URL from which to retrieve a file, the second value is the
        filepath where the file will be saved.
    auto_decompress: bool, optional, default True
        Automatically decompress responses.
    headers: LooseHeaders | None, default None
        Additional headers to send with each request.
    limit: int, optional, default 100
        Maximum number of simultaneous connections.
    chunk_size: int, optional, default 1024
        Amount of data to write at a time (KB).
    overwrite: bool, optional, default False
        If filepath exists, overwrite if True, else skip download.
    ssl_context: SSLContext, optional
        SSL configuration object. Uses system defaults unless otherwise
        specified.
    timeout: int, optional, default 300
        Maximum number of seconds to wait for a response to return.
    retries: int, optional, default 3
        Number of download attempts in the event of failure.
    
    Returns
    -------
    None
    """
    # SSL
    if ssl_context is None:
        ssl_context = ssl.create_default_context()

    # Retrieve
    connector = aiohttp.TCPConnector(limit=limit)
    async with aiohttp.ClientSession(
        connector=connector,
        headers=headers,
        auto_decompress=auto_decompress
        ) as session:
        await asyncio.gather(*(
            download_file_awaitable(
                url,
                filepath,
                session,
                chunk_size=chunk_size,
                overwrite=overwrite,
                ssl_context=ssl_context,
                timeout=timeout,
                retries=retries
                ) for (url, filepath) in src_dst
        ))

def download_files(
        *src_dst: tuple[str | URL, str | Path],
        auto_decompress: bool = True,
        headers: LooseHeaders | None = None,
        limit: int = 100,
        chunk_size: int = 1024,
        overwrite: bool = False,
        ssl_context: ssl.SSLContext | None = None,
        timeout: int = 300,
        file_validator: Callable[[Path], None] = default_file_validator,
        retries: int = 3
    ) -> None:
    """
    Asynchronously retrieve multiple files from urls and save to filepaths.
    
    Parameters
    ----------
    *src_dst: tuple[str | Path, str | URL], required
        One or more tuples containing two values. The first value is the 
        source URL from which to retrieve a file, the second value is the
        filepath where the file will be saved.
    auto_decompress: bool, optional, default True
        Automatically decompress responses.
    headers: LooseHeaders | None, default None
        Additional headers to send with each request.
    limit: int, optional, default 100
        Maximum number of simultaneous connections.
    chunk_size: int, optional, default 1024
        Amount of data to write at a time (KB).
    overwrite: bool, optional, default False
        If filepath exists, overwrite if True, else skip download.
    ssl_context: SSLContext, optional
        SSL configuration object. Uses system defaults unless otherwise
        specified.
    timeout: int, optional, default 300
        Maximum number of seconds to wait for a response to return.
    file_validator: Callable, optional
        Function used to validate files.
    retries: int, optional, default 3
        Number of download attempts in the event of failure.
    
    Returns
    -------
    None

    Examples
    --------
    >>> # This will download the pandas and numpy homepages and save them to 
    >>> # ./pandas_index.html and ./numpy_index.html
    >>> from downloads import download_files
    >>> download_files(
    ...     ("https://pandas.pydata.org/docs/user_guide/index.html", "pandas_index.html"),
    ...     ("https://numpy.org/doc/stable/index.html", "numpy_index.html")
    ...     )
    """
    # Get logger
    name = __loader__.name + "." + inspect.currentframe().f_code.co_name
    logger = get_logger(name)

    # SSL
    if ssl_context is None:
        ssl_context = ssl.create_default_context()

    # Retrieve
    for attempt in range(retries):
        logger.info("Downloading files, attempt %d", attempt)
        asyncio.run(
            download_files_awaitable(
                *src_dst,
                auto_decompress=auto_decompress,
                headers=headers,
                limit=limit,
                chunk_size=chunk_size,
                overwrite=overwrite,
                ssl_context=ssl_context,
                timeout=timeout,
                retries=retries
                )
            )

        # Validate files
        logger.info("Validating files, this will take some time")
        filepaths = [Path(dst) for _, dst in src_dst]
        validated = 0
        for fp in filepaths:
            logger.info("Validating %s", fp)
            if fp.exists():
                try:
                    file_validator(fp)
                    validated += 1
                except FileValidationError:
                    fp.unlink()
                    break
            else:
                break
        if validated == 0:
            warnings.warn("Unable to retrieve any files", RuntimeWarning)
        if validated == len(filepaths):
            logger.info("All files validated")
            return
        warnings.warn("Unable to retrieve all files, trying again", RuntimeWarning)
        sleep(5 * 2 ** attempt)
    warnings.warn("Unable to validate all files", RuntimeWarning)



"""Command-line Interface."""
from pathlib import Path
from datetime import datetime
from typing import Literal
from sys import stdout

import typer
import pandas as pd
import polars as pl

from nwm_explorer.configuration import load_configuration
from nwm_explorer.nwm import download_nwm
from nwm_explorer.usgs import download_usgs
from nwm_explorer.pairs import pair_nwm_usgs
from nwm_explorer.evaluate import evaluate as run_evaluation
from nwm_explorer.evaluate import scan_evaluations
from nwm_explorer.constants import ModelConfiguration, Metric, COLUMN_DESCRIPTIONS
from nwm_explorer.gui import serve_dashboards

app = typer.Typer()
"""Main typer command-line application."""

@app.command()
def download(
    start: datetime,
    end: datetime,
    root: Path,
    jobs: int = 1,
    retries: int = 3
    ) -> None:
    """
    Download and process NWM and USGS data for evaluations.
    """
    # Use time stamps
    start = pd.Timestamp(start)
    end = pd.Timestamp(end)

    # NWM downloads
    download_nwm(
        start=start,
        end=end,
        root=root,
        jobs=jobs,
        retries=retries
    )

    # USGS downloads
    download_usgs(
        start=start-pd.Timedelta("1d"),
        end=end+pd.Timedelta("10d"),
        root=root,
        retries=retries
    )

@app.command()
def pair(
    start: datetime,
    end: datetime,
    root: Path
    ) -> None:
    """
    Resample and pair NWM predictions to USGS observations.
    """
    # Use timestamps
    start = pd.Timestamp(start)
    end = pd.Timestamp(end)

    # Pair
    pair_nwm_usgs(
        root=root,
        start_date=start,
        end_date=end
    )

@app.command()
def compute(
    start: datetime,
    end: datetime,
    root: Path,
    label: str,
    jobs: int = 1,
    sites_per_chunk: int = 500
    ) -> None:
    """
    Compute evaluation metrics for NWM-USGS pairs.
    """
    # Use timestamps
    start = pd.Timestamp(start)
    end = pd.Timestamp(end)

    # Evaluate
    run_evaluation(
        label=label,
        root=root,
        start_time=start,
        end_time=end,
        processes=jobs,
        sites_per_chunk=sites_per_chunk
    )

@app.command()
def evaluate(configuration: Path = Path("config.json")) -> None:
    """
    Run standard evaluation including download, pair, and compute. Parameters
    set in configuration file.
    """
    # Load configuration
    config = load_configuration(configuration_file=configuration)

    # Process each evaluation
    for e in config.evaluations:
        # Set start and end times
        start = pd.Timestamp(e.start_time)
        end = pd.Timestamp(e.end_time)

        # NWM downloads
        download_nwm(
            start=start,
            end=end,
            root=config.root,
            jobs=config.processes,
            retries=config.retries
        )

        # USGS downloads
        download_usgs(
            start=start-pd.Timedelta("1d"),
            end=end+pd.Timedelta("10d"),
            root=config.root,
            retries=config.retries
        )

        # Pair
        pair_nwm_usgs(
            root=config.root,
            start_date=start,
            end_date=end
        )

        # Evaluate
        run_evaluation(
            label=e.label,
            root=config.root,
            start_time=start,
            end_time=end,
            processes=config.processes,
            sites_per_chunk=config.sites_per_chunk
        )

@app.command()
def export(
        root: Path,
        label: str,
        configuration: ModelConfiguration,
        metric: Metric,
        output: typer.FileTextWrite | None = None,
        lead_time_hours_min: int = 0,
        rank: Literal["min", "median", "max"] = "median",
        additional_columns: tuple[str] | None = None,
        no_data_value: str | None = None
    ) -> None:
    """
    Export evaluation metrics to CSV.
    """
    # Set output
    if output is None:
        output = stdout

    # Set additional columns
    if additional_columns is None:
        additional_columns = [
            "label",
            "configuration",
            "nwm_feature_id",
            "usgs_site_code",
            "reference_time_min",
            "reference_time_max",
            "lead_time_hours_min",
            "predicted_value_time_min",
            "predicted_value_time_max",
            f"predicted_cfs_{rank}",
            "observed_value_time_min",
            "observed_value_time_max",
            f"observed_cfs_{rank}",
            "sample_size"
        ]

    # Load metrics
    data = scan_evaluations(
        root
    ).filter(
        pl.col("label") == label,
        pl.col("configuration") == configuration,
        pl.col("lead_time_hours_min") == lead_time_hours_min
    ).select(
        additional_columns + [
            f"{metric}_{rank}_lower",
            f"{metric}_{rank}_point",
            f"{metric}_{rank}_upper"
        ]
    ).collect()

    # Write header
    header = "# National Water Model Evaluations\n# \n"
    for col in data.columns:
        header += f"# {col}: {COLUMN_DESCRIPTIONS.get(col)}\n"
    header += "# \n"
    output.write(header)

    # Write data
    output.write(data.write_csv(
        float_precision=2,
        datetime_format="%Y-%m-%dT%H:%M",
        null_value=no_data_value
    ))

@app.command()
def display(configuration: Path = Path("config.json")) -> None:
    """
    Launch graphical application in the browser. Shutdown the application using
    ctrl+c.
    """
    serve_dashboards(
        configuration_file=configuration
    )

def run() -> None:
    """Main entry point for CLI."""
    app()

if __name__ == "__main__":
    # Run the CLI.
    run()



"""Methods to pair predictions with observations."""
from pathlib import Path
import functools
import inspect

import pandas as pd
import polars as pl

from nwm_explorer.routelink import download_routelink
from nwm_explorer.nwm import scan_nwm
from nwm_explorer.usgs import scan_usgs, STATE_LIST
from nwm_explorer.logger import get_logger
from nwm_explorer.constants import (ModelConfiguration, LRU_CACHE_SIZES, SUBDIRECTORIES,
    GROUP_SPECIFICATIONS)

@functools.lru_cache(LRU_CACHE_SIZES["pairs"])
def routelink_cache(root: Path) -> pl.DataFrame:
    """Scan and collect routelink."""
    return download_routelink(root).select(
        ["nwm_feature_id", "usgs_site_code"]
    ).collect()

@functools.lru_cache(LRU_CACHE_SIZES["pairs"])
def observations_cache(root: Path) -> pl.LazyFrame:
    """Scan observations."""
    return scan_usgs(root)

@functools.lru_cache(LRU_CACHE_SIZES["pairs"])
def load_and_map_observations_chunk(
        root: Path,
        state_code: str,
        year: int,
        month: int
) -> pl.DataFrame:
    """
    Return polars.DataFrame of USGS observations. Add nwm_feature_id column.
    Cache DataFrame.

    Parameters
    ----------
    root: pathlib.Path
        Root data directory.
    start_time: str
        Start date. pandas.Timestamp compatible.
    end_time: str
        End date. pandas.Timestamp compatible.

    Returns
    -------
    polars.DataFrame
    """
    # Load routelink
    routelink = routelink_cache(root)

    # Scan observations
    observations = observations_cache(root)

    # Collect observations
    return observations.filter(
        pl.col("state_code") == state_code,
        pl.col("year") == year,
        pl.col("month") == month,
        pl.col("usgs_site_code").is_in(routelink["usgs_site_code"].implode())
    ).select(
        ["usgs_site_code", "value_time", "observed_cfs"]
    ).collect()

@functools.lru_cache(LRU_CACHE_SIZES["pairs"])
def load_and_map_observations(
        root: Path,
        start_time: str,
        end_time: str,
        window_interval: int
) -> pl.DataFrame:
    """
    Return polars.DataFrame of USGS observations. Add nwm_feature_id column.

    Parameters
    ----------
    root: pathlib.Path
        Root data directory.
    start_time: str
        Start date. pandas.Timestamp compatible.
    end_time: str
        End date. pandas.Timestamp compatible
    window_interval: int
        Time interval to aggregate over in hours.

    Returns
    -------
    polars.DataFrame
    """
    # Load routelink
    routelink = routelink_cache(root)

    # Expand range
    first = pd.Timestamp(start_time).to_period("M").start_time
    last = pd.Timestamp(end_time).to_period("M").start_time

    # Check end month
    if first == last:
        last += pd.Timedelta("31D")
        last = last.to_period("M").start_time

    # Load data
    dataframes = []
    for m in pd.date_range(first, last, freq="1MS"):
        for state_code in [s.abbr.lower() for s in STATE_LIST]:
            # Collect observations
            df = load_and_map_observations_chunk(
                root, state_code, m.year, m.month
            ).unique(
                ["usgs_site_code", "value_time"]
            ).sort(
                ["usgs_site_code", "value_time"]
            ).group_by_dynamic(
                "value_time",
                every=f"{window_interval}h",
                group_by="usgs_site_code"
            ).agg(
                pl.col("observed_cfs").min().alias("observed_cfs_min"),
                pl.col("observed_cfs").median().alias("observed_cfs_median"),
                pl.col("observed_cfs").max().alias("observed_cfs_max"),
                pl.col("value_time").min().alias("observed_value_time_min"),
                pl.col("value_time").max().alias("observed_value_time_max")
            ).with_columns(
                nwm_feature_id=pl.col("usgs_site_code").replace_strict(
                    routelink["usgs_site_code"].implode(),
                    routelink["nwm_feature_id"].implode()
                )
            )
            dataframes.append(df)

    # Concatenate
    return pl.concat(dataframes).filter(
        pl.col("value_time") >= start_time,
        pl.col("value_time") <= end_time
    )

def build_pairs_filepath(
    root: Path,
    configuration: ModelConfiguration,
    reference_date: pd.Timestamp
    ) -> Path:
    """Build and return hive-partitioned parquet file path."""
    config = f"configuration={configuration}"
    year = f"year={reference_date.year}"
    month = f"month={reference_date.month}"
    partition = f"P{reference_date.day}.parquet"
    subdirectory = SUBDIRECTORIES["pairs"]
    return root / subdirectory / config / year / month / partition

def pair_nwm_usgs(
        root: Path,
        start_date: pd.Timestamp,
        end_date: pd.Timestamp
        ) -> None:
    """
    Resample and pair NWM output to USGS observations.

    Parameters
    ----------
    root: pathlib.Path
        Root data directory.
    start_date: pandas.Timestamp
        First reference date to pair.
    end_date: int
        Last reference date to pair.
    """
    # Get logger
    name = __loader__.name + "." + inspect.currentframe().f_code.co_name
    logger = get_logger(name)

    # Pair
    prediction_store = scan_nwm(root)
    date_range = pd.date_range(
        start=start_date,
        end=end_date,
        freq="1MS"
    )
    logger.info("Pairing %s to %s", str(start_date), str(end_date))
    # Process each reference month
    for reference_date in date_range:
        # Process each configuration
        for config, specs in GROUP_SPECIFICATIONS.items():
            # Check for ofile
            ofile = build_pairs_filepath(
                root, config, reference_date
            )
            if ofile.exists():
                logger.info("Skipping existing: %s", ofile)
                continue
            logger.info("Building %s", ofile)

            # Aggregations
            aggregations = [
                pl.col("predicted_cfs").min().alias("predicted_cfs_min"),
                pl.col("predicted_cfs").median().alias("predicted_cfs_median"),
                pl.col("predicted_cfs").max().alias("predicted_cfs_max"),
                pl.col("value_time").min().alias("predicted_value_time_min"),
                pl.col("value_time").max().alias("predicted_value_time_max"),
            ]
            if "reference_time" not in specs.group_by_columns:
                aggregations.append(pl.col("reference_time").min())

            # Predictions
            logger.info("Resampling predictions")
            pred = prediction_store.filter(
                pl.col("configuration") == config,
                pl.col("year") == reference_date.year,
                pl.col("month") == reference_date.month
            ).select(
                specs.select_columns
            ).collect().sort(
                specs.sort_columns
            ).group_by_dynamic(
                specs.index_column,
                every=f"{specs.window_interval}h",
                group_by=specs.group_by_columns
            ).agg(*aggregations)

            # Get date range
            first = pred["value_time"].min()
            last = pred["value_time"].max()

            # Observations
            logger.info("Resampling observations")
            obs = load_and_map_observations(
                    root=root,
                    start_time=first,
                    end_time=last,
                    window_interval=specs.window_interval
                )

            # Pair and save
            logger.info("Pairing observations and predictions")
            ofile.parent.mkdir(exist_ok=True, parents=True)
            pairs = pred.join(obs, on=["nwm_feature_id", "value_time"],
                how="left").drop_nulls()

            # Save
            logger.info("Saving %s", ofile)
            pairs.write_parquet(ofile)

def scan_pairs_no_cache(root: Path) -> pl.LazyFrame:
    """
    Return polars.LazyFrame of paired NWM predictions and USGS observations.

    Parameters
    ----------
    root: pathlib.Path
        Root data directory.

    Returns
    -------
    polars.LazyFrame
    """
    # Get logger
    name = __loader__.name + "." + inspect.currentframe().f_code.co_name
    logger = get_logger(name)

    logger.info("Scanning pairs")
    subdirectory = SUBDIRECTORIES["pairs"]
    return pl.scan_parquet(
        root / f"{subdirectory}/",
        hive_schema={
            "configuration": pl.Enum(ModelConfiguration),
            "year": pl.Int32,
            "month": pl.Int32
        }
    )

@functools.lru_cache(LRU_CACHE_SIZES["pairs"])
def scan_pairs_cache(root: Path) -> pl.LazyFrame:
    """
    Return polars.LazyFrame of paired NWM predictions and USGS observations.
    Cache result.

    Parameters
    ----------
    root: pathlib.Path
        Root data directory.

    Returns
    -------
    polars.LazyFrame
    """
    return scan_pairs_no_cache(root)

def scan_pairs(root: Path, cache: bool = False) -> pl.LazyFrame:
    """
    Return polars.LazyFrame of paired NWM predictions and USGS observations.

    Parameters
    ----------
    root: pathlib.Path
        Root data directory.
    cache: bool, optional, default False
        Whether to cache the resulting LazyFrame for subsequent calls.

    Returns
    -------
    polars.LazyFrame
    """
    if cache:
        return scan_pairs_cache(root)
    return scan_pairs_no_cache(root)



"""Sidebar options that affect application behavior."""
from typing import Callable
from pathlib import Path

import panel as pn
from panel.viewable import Viewer

from nwm_explorer.constants import AxisType, MeasurementUnits
from nwm_explorer.usgs import load_site_information

UNIT_CONVERSIONS: dict[str, float] = {
    MeasurementUnits.CUBIC_FEET_PER_SECOND: 1.0,
    MeasurementUnits.CUBIC_METERS_PER_SECOND: 0.3048 ** 3.0,
    MeasurementUnits.CUBIC_FEET_PER_SECOND_PER_SQUARE_MILE: 1.0,
    MeasurementUnits.INCHES_PER_HOUR: 3.0 / 1936.0,
    MeasurementUnits.CUMULATIVE_INCHES_PER_HOUR: 3.0 / 1936.0
}
"""Measurement unit conversion functions. Assumes base units of CFS"""

def compute_conversion_factor(
        root: Path,
        usgs_site_code: str,
        measurement_units: str
) -> float:
    """
    Compute conversion factor for given site code and measurement units. Assumes
    source units are in cubic feet per second.
    """
    # Extract area
    if measurement_units in [
        MeasurementUnits.CUBIC_FEET_PER_SECOND_PER_SQUARE_MILE,
        MeasurementUnits.INCHES_PER_HOUR,
        MeasurementUnits.CUMULATIVE_INCHES_PER_HOUR]:
        # Load site information
        cols = ["drainage_area", "contributing_drainage_area"]
        df = load_site_information(root, usgs_site_code, cols)
        area = df.min_horizontal().item(0)
        if area is None:
            return 0.0
        return UNIT_CONVERSIONS[measurement_units] / area
    return UNIT_CONVERSIONS[measurement_units]

class StreamflowOptions(Viewer):
    """WidetBox for selecting streamflow plot parameters."""
    def __init__(self, **params):
        super().__init__(**params)

        # Setup
        self._widgets = {
            "axis_scale": pn.widgets.Select(
                name="Scale",
                options=["Linear", "Log"],
                value="Linear"
            ),
            "measurement_units": pn.widgets.Select(
                name="Measurement units",
                options=list(UNIT_CONVERSIONS.keys()),
                value=list(UNIT_CONVERSIONS.keys())[0]
            )
        }

    def __panel__(self):
        return pn.Column("# Streamflow options", *list(self._widgets.values()))

    def bind(self, function: Callable) -> None:
        """Bind function to widgets."""
        for k, w in self._widgets.items():
            pn.bind(function, w.param.value, watch=True, callback_type=k)

    @property
    def axis_scale(self) -> AxisType:
        """Currently select axis type."""
        return AxisType(self._widgets["axis_scale"].value.lower())

    @property
    def measurement_units(self) -> str:
        """Currently select measurment units."""
        return self._widgets["measurement_units"].value
